import unittest

from k8s_handle import settings
from k8s_handle.templating import get_template_contexts
from .mocks import K8sClientMock
from .mocks import ServiceMetadata
from .mocks import ServicePort
from .mocks import ServiceSpec
from .resource import Adapter
from .resource import Provisioner
from .resource import ProvisioningError


class TestProvisioner(unittest.TestCase):
    def setUp(self):
        settings.GET_ENVIRON_STRICT = False

    def test_deployment_wait_complete_fail(self):
        client = Adapter(
            api=K8sClientMock('test1'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'test1'}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', False, None)._wait_deployment_complete(client, tries=1, timeout=0)
        self.assertTrue('Deployment not completed for 1 tries' in str(context.exception), context.exception)

    def test_deployment_wait_complete(self):
        client = Adapter(
            api=K8sClientMock('test2'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'test1'}, 'spec': {'replicas': 1}})
        Provisioner('deploy', False, None)._wait_deployment_complete(client, tries=1, timeout=0)

    def test_statefulset_wait_complete_fail(self):
        client = Adapter(api=K8sClientMock('test1'),
                         spec={'kind': 'StatefulSet', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', False, None)._wait_statefulset_complete(client, tries=1, timeout=0)
        self.assertTrue('StatefulSet not completed for 1 tries' in str(context.exception), context.exception)

    def test_statefulset_wait_complete(self):
        client = Adapter(api=K8sClientMock('test2'),
                         spec={'kind': 'StatefulSet', 'metadata': {'name': ''}, 'spec': {'replicas': 3}})
        Provisioner('deploy', False, None)._wait_statefulset_complete(client, tries=1, timeout=0)

    def test_daemonset_wait_complete_fail(self):
        client = Adapter(api=K8sClientMock('test1'),
                         spec={'kind': 'DaemonSet', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', False, None)._wait_daemonset_complete(client, tries=1, timeout=0)
        self.assertTrue('DaemonSet not completed for 1 tries' in str(context.exception), context.exception)

    def test_daemonset_wait_complete(self):
        client = Adapter(api=K8sClientMock('test2'),
                         spec={'kind': 'DaemonSet', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        Provisioner('deploy', False, None)._wait_daemonset_complete(client, tries=1, timeout=0)

    def test_job_wait_complete_fail(self):
        client = Adapter(api=K8sClientMock('test1'),
                         spec={'kind': 'Job', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', False, None)._wait_job_complete(client, tries=1, timeout=0)

        self.assertTrue('Job running failed' in str(context.exception))

    def test_job_wait_complete_conditions_fail(self):
        client = Adapter(api=K8sClientMock('test2'),
                         spec={'kind': 'Job', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', False, None)._wait_job_complete(client, tries=1, timeout=0)
        self.assertTrue('Job not completed for 1 tries' in str(context.exception), context.exception)

    def test_job_wait_complete(self):
        client = Adapter(api=K8sClientMock('test3'),
                         spec={'kind': 'Job', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        Provisioner('deploy', False, None)._wait_job_complete(client, tries=1, timeout=0)

    def test_ns_from_template(self):
        client = Adapter(api=K8sClientMock('test'),
                         spec={'kind': 'Job', 'metadata': {'name': '', 'namespace': 'test'},
                               'spec': {'replicas': 1}})
        self.assertEqual(client.namespace, 'test')

    def test_ns_from_config(self):
        settings.K8S_NAMESPACE = 'namespace'
        client = Adapter(api=K8sClientMock('test'),
                         spec={'kind': 'Job', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        self.assertEqual(client.namespace, 'namespace')

    def test_deployment_destruction_wait_fail(self):
        client = Adapter(
            api=K8sClientMock('test1'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'test1'}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('destroy', False, None)._wait_destruction_complete(client, 'Deployment', tries=1, timeout=0)
        self.assertTrue('Deployment destruction not completed for 1 tries' in str(context.exception), context.exception)

    def test_deployment_destruction_wait_success(self):
        client = Adapter(
            api=K8sClientMock('404'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'test1'}, 'spec': {'replicas': 1}})
        Provisioner('destroy', False, None)._wait_destruction_complete(client, 'Deployment', tries=1, timeout=0)

    def test_job_destruction_wait_fail(self):
        client = Adapter(
            api=K8sClientMock('test1'),
            spec={'kind': 'Job', 'metadata': {'name': 'test1'}, 'spec': {'replicas': 1}})
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', True, None)._wait_destruction_complete(client, 'Job', tries=1, timeout=0)
        self.assertTrue('Job destruction not completed for 1 tries' in str(context.exception), context.exception)

    def test_job_destruction_wait_success(self):
        client = Adapter(
            api=K8sClientMock('404'),
            spec={'kind': 'Job', 'metadata': {'name': 'test1'}, 'spec': {'replicas': 1}})
        Provisioner('destroy', False, None)._wait_destruction_complete(client, 'Job', tries=1, timeout=0)

    def test_missing_annotations_and_labels(self):
        annotations = {
            'annotation1': 'value1',
            'annotation2': 'value2',
            'annotation3': 'value3',
            'kubectl.kubernetes.io/last-applied-configuration': {}
        }

        labels = {
            'label1': 'value1',
            'label2': 'value2',
            'label3': 'value3'
        }

        old_metadata = ServiceMetadata(annotations, labels)
        new_metadata = {
            'annotations': {'annotation1': 'value1', 'annotation2': 'value2'},
            'labels': {'label1': 'value1', 'label2': 'value2'}
        }

        missing_annotations, missing_labels = Provisioner('', False, None)._get_missing_annotations_and_labels(
            old_metadata, new_metadata)

        self.assertEqual(len(missing_annotations), 1)
        self.assertEqual(missing_annotations[0], 'annotation3')
        self.assertEqual(len(missing_labels), 1)
        self.assertEqual(missing_labels[0], 'label3')

    def test_no_missing_annotations_and_labels(self):
        annotations = {'annotation1': 'value1'}

        labels = {'label1': 'value1'}

        old_metadata = ServiceMetadata(annotations, labels)
        new_metadata = {
            'annotations': {'annotation1': 'value1', 'annotation2': 'value2'},
            'labels': {'label1': 'value1', 'label2': 'value2'}
        }

        missing_annotations, missing_labels = Provisioner('', False, None)._get_missing_annotations_and_labels(
            old_metadata, new_metadata)

        self.assertEqual(len(missing_annotations), 0)
        self.assertEqual(len(missing_labels), 0)

    def test_no_annotations_in_template(self):
        annotations = {'annotation1': 'value1', 'annotation2': 'value2'}

        labels = {'label1': 'value1'}

        old_metadata = ServiceMetadata(annotations, labels)
        new_metadata = {
            'labels': {'label1': 'value1', 'label2': 'value2'}
        }

        missing_annotations, missing_labels = Provisioner('', False, None)._get_missing_annotations_and_labels(
            old_metadata, new_metadata)

        self.assertEqual(len(missing_annotations), 2)
        self.assertEqual(len(missing_labels), 0)

    def test_no_annotations_in_service_metadata(self):
        labels = {'label1': 'value1'}

        old_metadata = ServiceMetadata(None, labels)
        new_metadata = {
            'annotations': {'annotation1': 'value1'},
            'labels': {'label1': 'value1', 'label2': 'value2'}
        }

        missing_annotations, missing_labels = Provisioner('', False, None)._get_missing_annotations_and_labels(
            old_metadata, new_metadata)

        self.assertEqual(len(missing_annotations), 0)
        self.assertEqual(len(missing_labels), 0)

    def test_missing_annotation_strict(self):
        missing_annotations = ['a1', 'a2', 'a3']
        settings.GET_ENVIRON_STRICT = True
        with self.assertRaises(RuntimeError) as context:
            Provisioner('', False, None)._notify_about_missing_items_in_template(items=missing_annotations,
                                                                                 missing_type='annotation')
        self.assertTrue('Please pay attention to service annotations!'
                        in str(context.exception))

    def test_missing_labels_strict(self):
        missing_labels = ['a1', 'a2', 'a3']
        settings.GET_ENVIRON_STRICT = True
        with self.assertRaises(RuntimeError) as context:
            Provisioner('', False, None)._notify_about_missing_items_in_template(items=missing_labels,
                                                                                 missing_type='label')
        self.assertTrue('Please pay attention to service labels!'
                        in str(context.exception))

    def test_missing_ports_strict(self):
        missing_ports = [ServicePort(80, 'test1'), ServicePort(90, 'test2'), ServicePort(50, None)]
        settings.GET_ENVIRON_STRICT = True
        with self.assertRaises(RuntimeError) as context:
            Provisioner('', False, None)._notify_about_missing_items_in_template(items=missing_ports,
                                                                                 missing_type='port')
        self.assertTrue('Please pay attention to service ports!'
                        in str(context.exception))

    def test_deploy_replace(self):
        settings.CHECK_STATUS_TIMEOUT = 0
        Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/deployment.yaml")

    def test_deploy_create(self):
        Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/deployment_404.yaml")

    def test_deploy_unknown_api(self):
        with self.assertRaises(RuntimeError) as context:
            Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/deployment_no_api.yaml")
        self.assertTrue('Unknown apiVersion "test" in template "k8s_handle/k8s/fixtures/deployment_no_api.yaml"'
                        in str(context.exception), context.exception)

    def test_service_replace(self):
        Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/service.yaml")

    def test_service_replace_no_ports(self):
        Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/service_no_ports.yaml")

    def test_destroy_unknown_api(self):
        with self.assertRaises(RuntimeError) as context:
            Provisioner('destroy', False, None).run("k8s_handle/k8s/fixtures/deployment_no_api.yaml")
        self.assertTrue('Unknown apiVersion "test" in template "k8s_handle/k8s/fixtures/deployment_no_api.yaml"'
                        in str(context.exception), context.exception)

    def test_destroy_not_found(self):
        Provisioner('destroy', False, None).run("k8s_handle/k8s/fixtures/deployment_404.yaml")

    def test_destroy_fail(self):
        with self.assertRaises(RuntimeError) as context:
            Provisioner('destroy', False, None).run("k8s_handle/k8s/fixtures/service.yaml")
        self.assertTrue('' in str(context.exception), context.exception)

    def test_destroy_success(self):
        Provisioner('destroy', False, None).run("k8s_handle/k8s/fixtures/deployment.yaml")

    def test_get_apply_ports_case1(self):
        # old_port=55, new_port=55, res: no apply ports
        old_spec = ServiceSpec('case1')
        new_spec = {'ports': [{'port': 55}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 0)

    def test_get_apply_ports_case2(self):
        # old_port=55, new_port=56, res: 1 apply port, 1 deleted
        old_spec = ServiceSpec('case2')
        new_spec = {'ports': [{'port': 56}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 2)
        self.assertDictEqual(apply_ports[0], {'port': 55, '$patch': 'delete'})
        self.assertDictEqual(apply_ports[1], {'port': 56})

    def test_get_apply_ports_case3(self):
        # old_port=55, new_port=56, name=test, res: 1 apply port, 1 deleted
        old_spec = ServiceSpec('case3')
        new_spec = {'ports': [{'port': 56, 'name': 'test'}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 2)
        self.assertDictEqual(apply_ports[0], {'port': 55, '$patch': 'delete'})
        self.assertDictEqual(apply_ports[1], {'port': 56, 'name': 'test'})

    def test_get_apply_ports_case4(self):
        # old_port=55,name=test, new_port=56,name=test, res: 1 apply port, 1 deleted
        old_spec = ServiceSpec('case4')
        new_spec = {'ports': [{'port': 56, 'name': 'foo'}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 2)
        self.assertDictEqual(apply_ports[0], {'port': 55, '$patch': 'delete'})
        self.assertDictEqual(apply_ports[1], {'port': 56, 'name': 'foo'})

    def test_get_apply_ports_case5(self):
        # add targetPort to existent port, res: 1 apply port
        old_spec = ServiceSpec('case5')
        new_spec = {'ports': [{'port': 55, 'name': 'test', 'targetPort': 90}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 1)
        self.assertDictEqual(apply_ports[0], {'port': 55, 'node_port': None,
                                              'protocol': 'TCP', 'name': 'test', 'targetPort': 90})

    def test_get_apply_ports_case6(self):
        # change targetPort, res: 1 apply port
        old_spec = ServiceSpec('case6')
        new_spec = {'ports': [{'port': 55, 'name': 'test', 'targetPort': 99}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 1)
        self.assertDictEqual(apply_ports[0], {'port': 55, 'node_port': None,
                                              'protocol': 'TCP', 'name': 'test', 'targetPort': 99})

    def test_get_apply_ports_case7(self):
        # change protocol, res: 1 apply port
        old_spec = ServiceSpec('case7')
        new_spec = {'ports': [{'port': 55, 'name': 'test', 'protocol': 'UDP'}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 1)
        self.assertDictEqual(apply_ports[0], {'port': 55, 'node_port': None,
                                              'protocol': 'UDP', 'name': 'test', 'target_port': 55})

    def test_get_apply_ports_case8(self):
        # add new ports, res: 1 delete, 2 apply
        old_spec = ServiceSpec('case8')
        new_spec = {'ports': [{'port': 90, 'name': 'test1'},
                              {'port': 99, 'name': 'test2'}]}
        apply_ports = Provisioner('deploy', False, None)._get_apply_ports(old_spec, new_spec)
        self.assertEqual(len(apply_ports), 3)
        self.assertDictEqual(apply_ports[0], {'port': 55, '$patch': 'delete'})
        self.assertDictEqual(apply_ports[1], {'port': 90, 'name': 'test1'})
        self.assertDictEqual(apply_ports[2], {'port': 99, 'name': 'test2'})

    def test_pvc_replace_equals(self):
        Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/pvc.yaml")

    def test_pvc_replace_not_equals(self):
        with self.assertRaises(ProvisioningError) as context:
            Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/pvc2.yaml")
        self.assertTrue('Replace persistent volume claim fail' in str(context.exception), context.exception)

    # https://kubernetes.io/docs/concepts/storage/persistent-volumes/#volume-mode
    def test_pvc_replace_new_attribute(self):
        with self.assertRaises(ProvisioningError) as context:
            Provisioner('deploy', False, None).run("k8s_handle/k8s/fixtures/pvc3.yaml")
        self.assertTrue('Replace persistent volume claim fail'
                        in str(context.exception))

    def test_get_template_contexts(self):
        with self.assertRaises(StopIteration):
            next(get_template_contexts('k8s_handle/k8s/fixtures/empty.yaml'))

        with self.assertRaises(RuntimeError) as context:
            next(get_template_contexts('k8s_handle/k8s/fixtures/nokind.yaml'))
        self.assertTrue(
            'Field "kind" not found (or empty) in file "k8s_handle/k8s/fixtures/nokind.yaml"' in str(context.exception),
            context.exception)

        with self.assertRaises(RuntimeError) as context:
            next(get_template_contexts('k8s_handle/k8s/fixtures/nometadata.yaml'))
        self.assertTrue(
            'Field "metadata" not found (or empty) in file "k8s_handle/k8s/fixtures/nometadata.yaml"'
            in str(context.exception),
            context.exception)

        with self.assertRaises(RuntimeError) as context:
            next(get_template_contexts('k8s_handle/k8s/fixtures/nometadataname.yaml'))
        self.assertTrue(
            'Field "metadata->name" not found (or empty) in file "k8s_handle/k8s/fixtures/nometadataname.yaml"'
            in str(context.exception), context.exception)

        context = next(get_template_contexts('k8s_handle/k8s/fixtures/valid.yaml'))
        self.assertEqual(context.get('kind'), 'Service')
        self.assertEqual(context.get('apiVersion'), 'v1')
        self.assertEqual(context.get('metadata').get('name'), 'my-service')
        self.assertEqual(context.get('spec').get('selector').get('app'), 'my-app')

        context = next(get_template_contexts('k8s_handle/k8s/fixtures/deployment_wo_replicas.yaml'))
        self.assertEqual(context.get('spec').get('replicas'), 1)


class TestKubeObject(unittest.TestCase):
    def test_replicas_equal(self):
        replicas = (1, 1, 1)
        self.assertTrue(Provisioner._replicas_count_are_greater_or_equal(replicas))

    def test_replicas_greater(self):
        replicas = (2, 3, 3)
        self.assertTrue(Provisioner._replicas_count_are_greater_or_equal(replicas))

    def test_replicas_not_equal(self):
        replicas = (1, 1, 0)
        self.assertFalse(Provisioner._replicas_count_are_greater_or_equal(replicas))

    def test_replicas_with_hpa_not_equal(self):
        replicas = (6, 7, 5)
        self.assertFalse(Provisioner._replicas_count_are_greater_or_equal(replicas))



import unittest
from .resource import Adapter
from .mocks import K8sClientMock

from .resource import ProvisioningError


class TestAdapter(unittest.TestCase):

    def test_get_app_kind(self):
        self.assertEqual(Adapter._get_app_kind('ConfigMap'), 'config_map')
        self.assertEqual(Adapter._get_app_kind('Namespace'), 'namespace')
        self.assertEqual(Adapter._get_app_kind('PodDisruptionBudget'), 'pod_disruption_budget')

    def test_app_kind_invalid(self):
        with self.assertRaises(RuntimeError) as context:
            Adapter(spec={'kind': 'UnknownKind', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        self.assertTrue('Unknown kind "UnknownKind" in generated file' in str(context.exception))

    def test_app_get_fail(self):
        deployment = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            deployment.get()
        self.assertTrue('Get deployment fail' in str(context.exception))

        storage = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'StorageClass', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            storage.get()
        self.assertTrue('Get storage class fail' in str(context.exception))

    def test_app_get_not_found(self):
        deployment = Adapter(
            api=K8sClientMock('404'),
            spec={'kind': 'Deployment', 'metadata': {'name': '404'}, 'spec': {'replicas': 1}})
        res = deployment.get()
        self.assertEqual(res, None)

        storage = Adapter(
            api=K8sClientMock('404'),
            spec={'kind': 'StorageClass', 'metadata': {'name': '404'}, 'spec': {'replicas': 1}})
        res = storage.get()
        self.assertEqual(res, None)

    def test_app_get(self):
        deployment = Adapter(
            api=K8sClientMock(),
            spec={'kind': 'Deployment', 'metadata': {'name': 'test'}, 'spec': {'replicas': 1}})
        res = deployment.get()

        self.assertEqual(res.metadata, {'key1': 'value1'})

        storage = Adapter(
            api=K8sClientMock(),
            spec={'kind': 'StorageClass', 'metadata': {'name': 'test'}, 'spec': {'replicas': 1}})
        res = storage.get()

        self.assertEqual(res.metadata, {'key1': 'value1'})

    def test_app_create_fail(self):
        deployment = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'Deployment', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            deployment.create()
        self.assertTrue('Create deployment fail' in str(context.exception))

        storage = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'StorageClass', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            storage.create()
        self.assertTrue('Create storage class fail' in str(context.exception))

    def test_app_create(self):
        deployment = Adapter(
            api=K8sClientMock(''),
            spec={'kind': 'Deployment', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        res = deployment.create()
        self.assertEqual(res, {'key1': 'value1'})

        storage = Adapter(
            api=K8sClientMock(''),
            spec={'kind': 'StorageClass', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        res = storage.create()
        self.assertEqual(res, {'key1': 'value1'})

    def test_app_replace_fail(self):
        deployment = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            deployment.replace()
        self.assertTrue('Replace deployment fail' in str(context.exception))

        storage = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'StorageClass', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            storage.replace()
        self.assertTrue('Replace storage class fail' in str(context.exception))

    def test_app_replace(self):
        deployment = Adapter(
            api=K8sClientMock(''),
            spec={'kind': 'Deployment', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        res = deployment.replace()
        self.assertEqual(res, {'key1': 'value1'})

        storage = Adapter(
            api=K8sClientMock(''),
            spec={'kind': 'StorageClass', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        res = storage.replace()
        self.assertEqual(res, {'key1': 'value1'})

    def test_app_replace_service(self):
        deployment = Adapter(
            api=K8sClientMock(''),
            spec={'kind': 'Service', 'metadata': {'name': ''}, 'spec': {'replicas': 1}})
        res = deployment.replace()
        self.assertEqual(res, {'key1': 'value1'})

    def test_app_delete_fail(self):
        client = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'Deployment', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            client.delete()
        self.assertTrue('Delete deployment fail' in str(context.exception))

        storage = Adapter(
            api=K8sClientMock('fail'),
            spec={'kind': 'StorageClass', 'metadata': {'name': 'fail'}, 'spec': {'replicas': 1}})
        with self.assertRaises(ProvisioningError) as context:
            storage.delete()
        self.assertTrue('Delete storage class fail' in str(context.exception))

    def test_app_delete_not_found(self):
        client = Adapter(
            api=K8sClientMock('404'),
            spec={'kind': 'Deployment', 'metadata': {'name': '404'}, 'spec': {'replicas': 1}})
        res = client.delete()
        self.assertEqual(res, None)

        storage = Adapter(
            api=K8sClientMock('404'),
            spec={'kind': 'StorageClass', 'metadata': {'name': '404'}, 'spec': {'replicas': 1}})
        res = storage.delete()
        self.assertEqual(res, None)

    def test_app_delete(self):
        client = Adapter(
            api=K8sClientMock(),
            spec={'kind': 'Deployment', 'metadata': {'name': 'test'}, 'spec': {'replicas': 1}})
        res = client.delete()

        self.assertEqual(res, {'key1': 'value1'})

        storage = Adapter(
            api=K8sClientMock(),
            spec={'kind': 'StorageClass', 'metadata': {'name': 'test'}, 'spec': {'replicas': 1}})
        res = storage.delete()

        self.assertEqual(res, {'key1': 'value1'})



import os
import glob
import base64
import logging
import yaml
from k8s_handle import settings
from hashlib import sha256
from jinja2 import Environment, FileSystemLoader, StrictUndefined
from jinja2.exceptions import TemplateNotFound, UndefinedError, TemplateSyntaxError


class TemplateRenderingError(Exception):
    pass


log = logging.getLogger(__name__)


def get_template_contexts(file_path):
    with open(file_path) as f:
        try:
            contexts = yaml.safe_load_all(f.read())
        except Exception as e:
            raise RuntimeError('Unable to load yaml file: {}, {}'.format(file_path, e))

        for context in contexts:
            if context is None:
                continue  # Skip empty YAML documents
            if 'kind' not in context or context['kind'] is None:
                raise RuntimeError('Field "kind" not found (or empty) in file "{}"'.format(file_path))
            if 'metadata' not in context or context['metadata'] is None:
                raise RuntimeError('Field "metadata" not found (or empty) in file "{}"'.format(file_path))
            if 'name' not in context['metadata'] or context['metadata']['name'] is None:
                raise RuntimeError('Field "metadata->name" not found (or empty) in file "{}"'.format(file_path))
            if 'spec' in context:
                # INFO: Set replicas = 1 by default for replaces cases in Deployment and StatefulSet
                if 'replicas' not in context['spec'] or context['spec']['replicas'] is None:
                    if context['kind'] in ['Deployment', 'StatefulSet']:
                        context['spec']['replicas'] = 1
            yield context


def b64decode(string):
    res = base64.decodebytes(string.encode())
    return res.decode()


def b64encode(string):
    res = base64.b64encode(string.encode())
    return res.decode()


def hash_sha256(string):
    res = sha256()
    res.update(string.encode('utf-8'))
    return res.hexdigest()


def get_env(templates_dir):
    # https://stackoverflow.com/questions/9767585/insert-static-files-literally-into-jinja-templates-without-parsing-them
    def include_file(path):
        path = os.path.join(templates_dir, '../', path)
        output = []
        for file_path in glob.glob(path):
            with open(file_path, 'r') as f:
                output.append(f.read())
        return '\n'.join(output)
    env = Environment(
        undefined=StrictUndefined,
        loader=FileSystemLoader([templates_dir]))

    env.filters['b64decode'] = b64decode
    env.filters['b64encode'] = b64encode
    env.filters['hash_sha256'] = hash_sha256
    env.globals['include_file'] = include_file

    log.debug('Available templates in path {}: {}'.format(templates_dir, env.list_templates()))
    return env


def _create_dir(dir):
    try:
        os.makedirs(dir)
    except os.error as e:
        log.debug(e)
        pass


class Renderer:
    def __init__(self, templates_dir):
        self._templates_dir = templates_dir
        self._env = get_env(self._templates_dir)

    def generate_by_context(self, context):
        if context is None:
            raise RuntimeError('Can\'t generate templates from None context')

        templates = context.get('templates', [])
        if len(templates) == 0:
            templates = context.get('kubectl', [])
            if len(templates) == 0:
                return

        output = []
        for template in templates:
            try:
                path = self._generate_file(template, settings.TEMP_DIR, context)
                log.info('File "{}" successfully generated'.format(path))
                output.append(path)
            except TemplateNotFound:
                raise TemplateRenderingError('Template "{}" not found'.format(template))
            except (UndefinedError, TemplateSyntaxError) as e:
                raise TemplateRenderingError('Unable to render {}, due to: {}'.format(template, e))
        return output

    def _generate_file(self, item, dir,  context):
        _create_dir(dir)
        try:
            log.info('Trying to generate file from template "{}" in "{}"'.format(item['template'], dir))
            template = self._env.get_template(item['template'])
        except TemplateNotFound as e:
            log.info('Templates path: {}, available templates:{}'.format(self._templates_dir,
                                                                         self._env.list_templates()))
            raise e
        except KeyError:
            raise RuntimeError('Templates section doesn\'t have any template items')
        new_name = item['template'].replace('.j2', '')
        path = os.path.join(dir, new_name)
        if not os.path.exists(os.path.dirname(path)):
            _create_dir(os.path.dirname(path))
        with open(path, 'w+') as f:
            f.write(template.render(context))
        return path



from kubernetes.client.rest import ApiException
from collections import namedtuple


class K8sClientMock:
    def __init__(self, name=None):
        self.name = name
        pass

    # Deployment
    def read_namespaced_deployment(self, name, namespace):
        if self.name == 'fail':
            raise ApiException('Get deployment fail')
        if self.name == '404' or name == '404':
            raise ApiException(reason='Not Found')

        my_response = namedtuple('my_response', 'metadata status')
        my_status = namedtuple('my_status',
                               'replicas available_replicas ready_replicas updated_replicas unavailable_replicas')

        if self.name == 'test1':
            return my_response(metadata={}, status=my_status(replicas=3,
                                                             available_replicas=2,
                                                             ready_replicas=1,
                                                             updated_replicas=None,
                                                             unavailable_replicas=1))
        if self.name == 'test2' or name == 'test2':
            return my_response(metadata={}, status=my_status(replicas=1,
                                                             available_replicas=1,
                                                             ready_replicas=1,
                                                             updated_replicas=1,
                                                             unavailable_replicas=None))

        return my_response(metadata={'key1': 'value1'}, status={'key1': 'value1'})

    def create_namespaced_deployment(self, body, namespace):
        if self.name == 'fail':
            raise ApiException('Create deployment fail')

        return {'key1': 'value1'}

    def replace_namespaced_deployment(self, name, body, namespace):
        if self.name == 'fail':
            raise ApiException('Replace deployment fail')

        return {'key1': 'value1'}

    def delete_namespaced_deployment(self, name, body, namespace):
        if self.name == 'fail':
            raise ApiException('Delete deployment fail')
        if self.name == '404' or name == '404':
            raise ApiException(reason='Not Found')

        if self.name == 'test1' or name == 'test1':
            my_response = namedtuple('my_response', 'message')
            return my_response(message='Failed')

        if self.name == 'test2' or name == 'test2':
            my_response = namedtuple('my_response', 'message')
            return my_response(message=None)

        return {'key1': 'value1'}

    # Service
    def read_namespaced_service(self, name, namespace, body=None):
        if self.name == 'fail':
            raise ApiException('Get service fail')
        if self.name == '404':
            raise ApiException(reason='Not Found')

        my_response = namedtuple('my_response', 'metadata status spec')
        my_status = namedtuple('my_status',
                               'replicas available_replicas ready_replicas updated_replicas unavailable_replicas')
        my_spec = namedtuple('my_spec', 'ports')
        my_port = namedtuple('my_port', 'port name')

        if self.name == 'test1':
            return my_response(metadata={}, spec=my_spec(ports=[my_port(port=123, name='test1')]),
                               status=my_status(replicas=3,
                                                available_replicas=2,
                                                ready_replicas=1,
                                                updated_replicas=None,
                                                unavailable_replicas=1))
        if self.name == 'test2':
            return my_response(metadata={}, spec=my_spec(ports=[]),
                               status=my_status(replicas=1,
                                                available_replicas=1,
                                                ready_replicas=1,
                                                updated_replicas=1,
                                                unavailable_replicas=None))

        return my_response(metadata={'key1': 'value1'}, status={'key1': 'value1'}, spec={})

    def replace_namespaced_service(self, name, body, namespace):
        if self.name == 'fail':
            raise ApiException('Replace service fail')

        return {'key1': 'value1'}

    def delete_namespaced_service(self, name, namespace, body=None):
        my_response = namedtuple('my_response', 'message')
        return my_response(message='Failed')

    def patch_namespaced_service(self, name, body, namespace):
        return {'key1': 'value1'}

    # StatefulSet
    def read_namespaced_stateful_set(self, name, namespace):
        if self.name == 'fail':
            raise ApiException('Get statefulset fail')
        if self.name == '404':
            raise ApiException(reason='Not Found')

        my_response = namedtuple('my_response', 'metadata status')
        my_status = namedtuple('my_status',
                               'current_replicas current_revision ready_replicas replicas update_revision')

        if self.name == 'test1':
            return my_response(metadata={}, status=my_status(current_replicas=2,
                                                             current_revision='revision-123',
                                                             ready_replicas=1,
                                                             replicas=3,
                                                             update_revision='revision-321'))

        if self.name == 'test2':
            return my_response(metadata={}, status=my_status(current_replicas=3,
                                                             current_revision='revision-123',
                                                             ready_replicas=3,
                                                             replicas=3,
                                                             update_revision='revision-123'))

        return my_response(metadata={'key1': 'value1'}, status={'key1': 'value1'})

    # DaemonSet
    def read_namespaced_daemon_set(self, name, namespace):
        if self.name == 'fail':
            raise ApiException('Get daemonset fail')
        if self.name == '404':
            raise ApiException(reason='Not Found')

        my_response = namedtuple('my_response', 'metadata status')
        my_status = namedtuple('my_status', 'desired_number_scheduled number_available '
                                            'number_ready updated_number_scheduled number_unavailable')

        if self.name == 'test1':
            return my_response(metadata={}, status=my_status(desired_number_scheduled=2,
                                                             number_available=2,
                                                             number_ready=1,
                                                             updated_number_scheduled=1,
                                                             number_unavailable=1))
        if self.name == 'test2':
            return my_response(metadata={}, status=my_status(desired_number_scheduled=2,
                                                             number_available=2,
                                                             number_ready=2,
                                                             updated_number_scheduled=2,
                                                             number_unavailable=None))

        return my_response(metadata={'key1': 'value1'}, status={'key1': 'value1'})

    # Job
    def read_namespaced_job(self, name, namespace):
        if self.name == 'fail':
            raise ApiException('Get daemonset fail')
        if self.name == '404':
            raise ApiException(reason='Not Found')

        my_response = namedtuple('my_response', 'metadata status')
        my_status = namedtuple('my_status', 'failed conditions')

        if self.name == 'test1':
            return my_response(metadata={}, status=my_status(failed='Failed',
                                                             conditions=[]))
        if self.name == 'test2':
            my_conditions = namedtuple('my_conditions', 'type')
            return my_response(metadata={}, status=my_status(failed=None,
                                                             conditions=[my_conditions(type='Failed')]))
        if self.name == 'test3':
            my_conditions = namedtuple('my_conditions', 'type')
            return my_response(metadata={}, status=my_status(failed=None,
                                                             conditions=[my_conditions(type='Complete')]))

        return my_response(metadata={'key1': 'value1'}, status={'key1': 'value1'})

    # StorageClass
    def read_storage_class(self, name):
        if self.name == 'fail':
            raise ApiException('Get storage class fail')
        if self.name == '404' or name == '404':
            raise ApiException(reason='Not Found')

        my_response = namedtuple('my_response', 'metadata status')
        my_status = namedtuple('my_status',
                               'replicas available_replicas ready_replicas updated_replicas unavailable_replicas')

        if self.name == 'test1':
            return my_response(metadata={}, status=my_status(replicas=3,
                                                             available_replicas=2,
                                                             ready_replicas=1,
                                                             updated_replicas=None,
                                                             unavailable_replicas=1))
        if self.name == 'test2' or name == 'test2':
            return my_response(metadata={}, status=my_status(replicas=1,
                                                             available_replicas=1,
                                                             ready_replicas=1,
                                                             updated_replicas=1,
                                                             unavailable_replicas=None))

        return my_response(metadata={'key1': 'value1'}, status={'key1': 'value1'})

    def create_storage_class(self, body):
        if self.name == 'fail':
            raise ApiException('Create storage class fail')

        return {'key1': 'value1'}

    def replace_storage_class(self, name, body):
        if self.name == 'fail':
            raise ApiException('Replace storage class fail')

        return {'key1': 'value1'}

    def delete_storage_class(self, name, body):
        if self.name == 'fail':
            raise ApiException('Delete storage class fail')
        if self.name == '404' or name == '404':
            raise ApiException(reason='Not Found')

        if self.name == 'test1' or name == 'test1':
            my_response = namedtuple('my_response', 'message')
            return my_response(message='Failed')

        if self.name == 'test2' or name == 'test2':
            my_response = namedtuple('my_response', 'message')
            return my_response(message=None)

        return {'key1': 'value1'}

    # PersistentVolumeClaim
    def read_namespaced_persistent_volume_claim(self, name, namespace):
        my_response = namedtuple('my_response', 'spec metadata')
        my_spec = namedtuple('my_spec',
                             'access_modes resources selector storage_class_name')

        if self.name == 'test1' or name == 'test1':
            return my_response(metadata={}, spec=my_spec(access_modes=['ReadWriteOnce'],
                                                         resources={'requests': {'storage': '1Gi'}},
                                                         storage_class_name='test',
                                                         selector={'matchLabels': {'volume': 'test'}}))

        if self.name == 'test2' or name == 'test2':
            return my_response(metadata={}, spec=my_spec(access_modes=['ReadWriteOnce'],
                                                         resources={'requests': {'storage': '2Gi'}},
                                                         storage_class_name='test',
                                                         selector={'matchLabels': {'volume': 'test'}}))

        return my_response(metadata={'key1': 'value1'}, spec={'key1': 'value1'})

    def replace_namespaced_persistent_volume_claim(self, name, body, namespace):
        if self.name == 'test2' or name == 'test2':
            raise ApiException('Replace persistent volume claim fail')


class ServiceMetadata:
    labels = {}

    def __init__(self, annotations, labels):
        if annotations is not None:
            self.annotations = annotations
        self.labels = labels


class ServiceSpec:
    def __init__(self, case):
        if case in ['case1', 'case2', 'case3']:
            self.ports = [ServicePort(55)]
        if case in ['case4', 'case5', 'case7', 'case8']:
            self.ports = [ServicePort(55, 'test')]
        if case in ['case6']:
            self.ports = [ServicePort(55, 'test', 90)]


class ServicePort:
    def __init__(self, port, name=None, target_port=None, node_port=None, protocol='TCP'):
        self.port = port
        self.name = name
        self.node_port = node_port
        self.protocol = protocol
        if target_port is None:
            self.target_port = port
        else:
            self.target_port = port
